import pickle
from io import BytesIO
from unittest.mock import Mock, patch

import numpy as np
import pandas as pd

from rdt.transformers import (
    CategoricalTransformer, LabelEncodingTransformer, OneHotEncodingTransformer)


def test_categorical_numerical_nans():
    """Ensure CategoricalTransformer works on numerical + nan only columns."""

    data = pd.DataFrame([1, 2, float('nan'), np.nan], columns=['column_name'])

    transformer = CategoricalTransformer()
    transformer.fit(data, list(data.columns))
    transformed = transformer.transform(data)
    reverse = transformer.reverse_transform(transformed)

    pd.testing.assert_frame_equal(reverse, data)


def test_categoricaltransformer_pickle_nans():
    """Ensure that CategoricalTransformer can be pickled and loaded with nan value."""
    # setup
    data = pd.DataFrame([1, 2, float('nan'), np.nan], columns=['column_name'])

    transformer = CategoricalTransformer()
    transformer.fit(data, list(data.columns))
    transformed = transformer.transform(data)

    # create pickle file on memory
    bytes_io = BytesIO()
    pickle.dump(transformer, bytes_io)
    # rewind
    bytes_io.seek(0)

    # run
    pickled_transformer = pickle.load(bytes_io)

    # assert
    pickle_transformed = pickled_transformer.transform(data)
    pd.testing.assert_frame_equal(pickle_transformed, transformed)


def test_categoricaltransformer_strings():
    """Test the CategoricalTransformer on string data.

    Ensure that the CategoricalTransformer can fit, transform, and reverse
    transform on string data. Expect that the reverse transformed data
    is the same as the input.

    Input:
        - 4 rows of string data
    Output:
        - The reverse transformed data
    """
    # setup
    data = pd.DataFrame(['a', 'b', 'a', 'c'], columns=['column_name'])
    transformer = CategoricalTransformer()

    # run
    transformer.fit(data, list(data.columns))
    reverse = transformer.reverse_transform(transformer.transform(data))

    # assert
    pd.testing.assert_frame_equal(data, reverse)


def test_categoricaltransformer_strings_2_categories():
    """Test the CategoricalTransformer on string data.

    Ensure that the CategoricalTransformer can fit, transform, and reverse
    transform on string data, when there are 2 categories of strings with
    the same value counts. Expect that the reverse transformed data is the
    same as the input.

    Input:
        - 4 rows of string data
    Output:
        - The reverse transformed data
    """
    # setup
    data = pd.DataFrame(['a', 'b', 'a', 'b'], columns=['column_name'])
    transformer = CategoricalTransformer()

    transformer.fit(data, list(data.columns))
    reverse = transformer.reverse_transform(transformer.transform(data))

    # assert
    pd.testing.assert_frame_equal(data, reverse)


def test_categoricaltransformer_integers():
    """Test the CategoricalTransformer on integer data.

    Ensure that the CategoricalTransformer can fit, transform, and reverse
    transform on integer data. Expect that the reverse transformed data is the
    same as the input.

    Input:
        - 4 rows of int data
    Output:
        - The reverse transformed data
    """
    # setup
    data = pd.DataFrame([1, 2, 3, 2], columns=['column_name'])
    transformer = CategoricalTransformer()

    # run
    transformer.fit(data, list(data.columns))
    reverse = transformer.reverse_transform(transformer.transform(data))

    # assert
    pd.testing.assert_frame_equal(data, reverse)


def test_categoricaltransformer_bool():
    """Test the CategoricalTransformer on boolean data.

    Ensure that the CategoricalTransformer can fit, transform, and reverse
    transform on boolean data. Expect that the reverse transformed data is the
    same as the input.

    Input:
        - 4 rows of bool data
    Output:
        - The reverse transformed data
    """
    # setup
    data = pd.DataFrame([True, False, True, False], columns=['column_name'])
    transformer = CategoricalTransformer()

    # run
    transformer.fit(data, list(data.columns))
    reverse = transformer.reverse_transform(transformer.transform(data))

    # assert
    pd.testing.assert_frame_equal(data, reverse)


def test_categoricaltransformer_mixed():
    """Test the CategoricalTransformer on mixed type data.

    Ensure that the CategoricalTransformer can fit, transform, and reverse
    transform on mixed type data. Expect that the reverse transformed data is
    the same as the input.

    Input:
        - 4 rows of mixed data
    Output:
        - The reverse transformed data
    """
    # setup
    data = pd.DataFrame([True, 'a', 1, None], columns=['column_name'])
    transformer = CategoricalTransformer()

    # run
    transformer.fit(data, list(data.columns))
    reverse = transformer.reverse_transform(transformer.transform(data))

    # assert
    pd.testing.assert_frame_equal(data, reverse)


@patch('psutil.virtual_memory')
def test_categoricaltransformer_mixed_low_virtual_memory(psutil_mock):
    """Test the CategoricalTransformer on mixed type data with low virtual memory.

    Ensure that the CategoricalTransformer can fit, transform, and reverse
    transform on mixed type data, when there is low virtual memory. Expect that the
    reverse transformed data is the same as the input.

    Input:
        - 4 rows of mixed data
    Output:
        - The reverse transformed data
    """
    # setup
    data = pd.DataFrame([True, 'a', 1, None], columns=['column_name'])
    transformer = CategoricalTransformer()

    virtual_memory = Mock()
    virtual_memory.available = 1
    psutil_mock.return_value = virtual_memory

    # run
    transformer.fit(data, list(data.columns))
    reverse = transformer.reverse_transform(transformer.transform(data))

    # assert
    pd.testing.assert_frame_equal(data, reverse)


@patch('psutil.virtual_memory')
def test_categoricaltransformer_mixed_more_rows(psutil_mock):
    """Test the CategoricalTransformer on mixed type data with low virtual memory.

    Ensure that the CategoricalTransformer can fit, transform, and reverse
    transform on mixed type data, when there is low virtual memory and a larger
    number of rows. Expect that the reverse transformed data is the same as the input.

    Input:
        - 4 rows of mixed data
    Output:
        - The reverse transformed data
    """
    # setup
    data = pd.DataFrame([True, 'a', 1, None], columns=['column_name'])
    transform_data = pd.DataFrame(['a', 1, None, 'a', True, 1], columns=['column_name'])
    transformer = CategoricalTransformer()

    virtual_memory = Mock()
    virtual_memory.available = 1
    psutil_mock.return_value = virtual_memory

    # run
    transformer.fit(data, list(data.columns))
    transformed = transformer.transform(transform_data)
    reverse = transformer.reverse_transform(transformed)

    # assert
    pd.testing.assert_frame_equal(transform_data, reverse)


def test_one_hot_numerical_nans():
    """Ensure OneHotEncodingTransformer works on numerical + nan only columns."""

    data = pd.DataFrame([1, 2, float('nan'), np.nan], columns=['column_name'])

    transformer = OneHotEncodingTransformer()
    transformer.fit(data, list(data.columns))
    transformed = transformer.transform(data)
    reverse = transformer.reverse_transform(transformed)

    pd.testing.assert_frame_equal(reverse, data)


def test_label_numerical_2d_array():
    """Ensure LabelEncodingTransformer works on numerical + nan only columns."""

    data = pd.DataFrame(['a', 'b', 'c', 'd'], columns=['column_name'])

    transformer = LabelEncodingTransformer()
    transformer.fit(data, list(data.columns))
    transformed = pd.DataFrame([0, 1, 2, 3], columns=['column_name.value'])
    reverse = transformer.reverse_transform(transformed)

    pd.testing.assert_frame_equal(reverse, data)


def test_label_numerical_nans():
    """Ensure LabelEncodingTransformer works on numerical + nan only columns."""

    data = pd.DataFrame([1, 2, float('nan'), np.nan], columns=['column_name'])

    transformer = LabelEncodingTransformer()
    transformer.fit(data, list(data.columns))
    transformed = transformer.transform(data)
    reverse = transformer.reverse_transform(transformed)

    pd.testing.assert_frame_equal(reverse, data)



"""Integration tests for the HyperTransformer."""

from copy import deepcopy
from unittest.mock import patch

import numpy as np
import pandas as pd

from rdt import HyperTransformer
from rdt.transformers import (
    DEFAULT_TRANSFORMERS, BaseTransformer, BooleanTransformer, CategoricalTransformer,
    DatetimeTransformer, NumericalTransformer, OneHotEncodingTransformer)


class DummyTransformerNumerical(BaseTransformer):

    INPUT_TYPE = 'categorical'
    OUTPUT_TYPES = {
        'value': 'float'
    }

    def _fit(self, data):
        pass

    def _transform(self, data):
        return data.astype(float)

    def _reverse_transform(self, data):
        return data.astype(str)


class DummyTransformerNotMLReady(BaseTransformer):

    INPUT_TYPE = 'datetime'
    OUTPUT_TYPES = {
        'value': 'categorical',
    }

    def _fit(self, data):
        pass

    def _transform(self, data):
        # Stringify input data
        return data.astype(str)

    def _reverse_transform(self, data):
        return data.astype('datetime64')


class DummyTransformerMultiColumn(BaseTransformer):

    INPUT_TYPE = 'datetime'
    OUTPUT_TYPES = {
        'value': 'float',
    }

    def _fit(self, data):
        pass

    def _transform(self, data):
        # Convert multiple columns into a single datetime
        data.columns = [c.replace('_str.value', '') for c in data.columns]
        data = pd.to_datetime(data)

        float_data = data.to_numpy().astype(np.float64)
        data_is_nan = data.isna().to_numpy().astype(np.float64)

        output = dict(zip(
            self.output_columns,
            [float_data, data_is_nan]
        ))

        output = pd.DataFrame(output).fillna(-1)

        return output

    def _reverse_transform(self, data):
        datetimes = data.round().astype('datetime64[ns]')
        out = pd.DataFrame({
            'year': datetimes.dt.year,
            'month': datetimes.dt.month,
            'day': datetimes.dt.day,
        })

        return out


TEST_DATA_INDEX = [4, 6, 3, 8, 'a', 1.0, 2.0, 3.0]


def get_input_data():
    datetimes = pd.to_datetime([
        np.nan,
        '2010-02-01',
        '2010-01-01',
        '2010-01-01',
        '2010-01-01',
        '2010-02-01',
        '2010-01-01',
        '2010-01-01',
    ])
    data = pd.DataFrame({
        'integer': [1, 2, 1, 3, 1, 4, 2, 3],
        'float': [0.1, 0.2, 0.1, np.nan, 0.1, 0.4, np.nan, 0.3],
        'categorical': ['a', 'a', np.nan, 'b', 'a', 'b', 'a', 'a'],
        'bool': [False, np.nan, False, True, False, True, True, False],
        'datetime': datetimes,
        'names': ['Jon', 'Arya', 'Arya', 'Jon', 'Jon', 'Sansa', 'Jon', 'Jon'],
    }, index=TEST_DATA_INDEX)

    return data


def get_transformed_data():
    datetimes = [
        1.263069e+18,
        1.264982e+18,
        1.262304e+18,
        1.262304e+18,
        1.262304e+18,
        1.264982e+18,
        1.262304e+18,
        1.262304e+18
    ]
    return pd.DataFrame({
        'integer.value': [1, 2, 1, 3, 1, 4, 2, 3],
        'float.value': [0.1, 0.2, 0.1, 0.2, 0.1, 0.4, 0.2, 0.3],
        'float.is_null': [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0],
        'categorical.value': [0.3125, 0.3125, 0.9375, 0.75, 0.3125, 0.75, 0.3125, 0.3125],
        'bool.value': [0.0, -1.0, 0.0, 1.0, 0.0, 1.0, 1.0, 0.0],
        'bool.is_null': [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        'datetime.value': datetimes,
        'datetime.is_null': [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        'names.value': [0.3125, 0.75, 0.75, 0.3125, 0.3125, 0.9375, 0.3125, 0.3125]
    }, index=TEST_DATA_INDEX)


DETERMINISTIC_DEFAULT_TRANSFORMERS = deepcopy(DEFAULT_TRANSFORMERS)
DETERMINISTIC_DEFAULT_TRANSFORMERS['categorical'] = CategoricalTransformer


@patch('rdt.transformers.DEFAULT_TRANSFORMERS', DETERMINISTIC_DEFAULT_TRANSFORMERS)
def test_hypertransformer_default_inputs():
    """Test the HyperTransformer with default parameters.

    This tests that if default parameters are provided to the HyperTransformer,
    the ``default_transformers`` method will be used to determine which
    transformers to use for each field.

    Setup:
        - Patch the DEFAULT_TRANSFORMERS to use the `CategoricalTransformer`
        for categorical data types, so that the output is predictable.

    Input:
        - A dataframe with every data type.

    Expected behavior:
        - The transformed data should contain all the ML ready data.
        - The reverse transformed data should be the same as the input.
    """
    # Setup
    data = get_input_data()

    # Run
    ht = HyperTransformer()
    ht.fit(data)
    transformed = ht.transform(data)
    reverse_transformed = ht.reverse_transform(transformed)

    # Assert
    expected_transformed = get_transformed_data()
    pd.testing.assert_frame_equal(transformed, expected_transformed)

    expected_reversed = get_input_data()
    pd.testing.assert_frame_equal(expected_reversed, reverse_transformed)

    assert isinstance(ht._transformers_tree['integer']['transformer'], NumericalTransformer)
    assert ht._transformers_tree['integer']['outputs'] == ['integer.value']
    assert isinstance(ht._transformers_tree['float']['transformer'], NumericalTransformer)
    assert ht._transformers_tree['float']['outputs'] == ['float.value', 'float.is_null']
    assert isinstance(ht._transformers_tree['categorical']['transformer'], CategoricalTransformer)
    assert ht._transformers_tree['categorical']['outputs'] == ['categorical.value']
    assert isinstance(ht._transformers_tree['bool']['transformer'], BooleanTransformer)
    assert ht._transformers_tree['bool']['outputs'] == ['bool.value', 'bool.is_null']
    assert isinstance(ht._transformers_tree['datetime']['transformer'], DatetimeTransformer)
    assert ht._transformers_tree['datetime']['outputs'] == ['datetime.value', 'datetime.is_null']
    assert isinstance(ht._transformers_tree['names']['transformer'], CategoricalTransformer)
    assert ht._transformers_tree['names']['outputs'] == ['names.value']


def test_hypertransformer_field_transformers():
    """Test the HyperTransformer with ``field_transformers`` provided.

    This tests that this transformers specified in the ``field_transformers``
    argument are used. Any output of a transformer that is not ML ready (not
    in the ``_transform_output_types`` list) should be recursively transformed
    till it is.

    Setup:
        - The datetime column is set to use a dummy transformer that stringifies
        the input. That output is then set to use the categorical transformer.

    Input:
        - A dict mapping each field to a transformer.
        - A dataframe with every data type.

    Expected behavior:
        - The transformed data should contain all the ML ready data.
        - The reverse transformed data should be the same as the input.
    """
    # Setup
    field_transformers = {
        'integer': NumericalTransformer(dtype=np.int64),
        'float': NumericalTransformer(dtype=float),
        'categorical': CategoricalTransformer,
        'bool': BooleanTransformer,
        'datetime': DummyTransformerNotMLReady,
        'datetime.value': CategoricalTransformer,
        'names': CategoricalTransformer
    }
    data = get_input_data()

    # Run
    ht = HyperTransformer(field_transformers=field_transformers)
    ht.fit(data)
    transformed = ht.transform(data)
    reverse_transformed = ht.reverse_transform(transformed)

    # Assert
    expected_transformed = get_transformed_data()
    del expected_transformed['datetime.is_null']
    rename = {'datetime.value': 'datetime.value.value'}
    expected_transformed = expected_transformed.rename(columns=rename)
    transformed_datetimes = [0.9375, 0.75, 0.3125, 0.3125, 0.3125, 0.75, 0.3125, 0.3125]
    expected_transformed['datetime.value.value'] = transformed_datetimes
    pd.testing.assert_frame_equal(transformed, expected_transformed)

    expected_reversed = get_input_data()
    pd.testing.assert_frame_equal(expected_reversed, reverse_transformed)


def test_hypertransformer_field_transformers_multi_column_fields():
    """Test the HyperTransformer with ``field_transformers`` provided.

    This test will make sure that fields made up of multiple columns are
    properly handled if they are specified in the ``field_transformers``
    argument. If the field has one column that is derived from another
    transformer, then the other transformer should be transformed and
    the multi-column field should be handled when all its columns are
    present.

    Setup:
        - A dummy transformer that takes in a column for day, year and
        month and creates one numeric value from it.
        - A dummy transformer that takes a string value and parses the
        float from it.

    Input:
        - A dict mapping each field to a dummy transformer.
        - A dataframe with a numerical year, month and day column as well
        as a string year, month and day column.

    Expected behavior:
        - The transformed data should contain all the ML ready data.
        - The reverse transformed data should be the same as the input.
    """
    # Setup
    data = pd.DataFrame({
        'year': [2001, 2002, 2003],
        'month': [1, 2, 3],
        'day': [1, 2, 3],
        'year_str': ['2001', '2002', '2003'],
        'month_str': ['1', '2', '3'],
        'day_str': ['1', '2', '3'],
    })
    field_transformers = {
        ('year', 'month', 'day'): DummyTransformerMultiColumn,
        'year_str': DummyTransformerNumerical,
        'day_str': DummyTransformerNumerical,
        'month_str': DummyTransformerNumerical,
        ('year_str.value', 'month_str.value', 'day_str.value'): DummyTransformerMultiColumn
    }
    expected_transformed = pd.DataFrame({
        'year#month#day.value': [
            9.783072e+17,
            1.012608e+18,
            1.046650e+18
        ],
        'year_str.value#month_str.value#day_str.value.value': [
            9.783072e+17,
            1.012608e+18,
            1.046650e+18
        ]
    })
    expected_reversed = data.copy()

    # Run
    ht = HyperTransformer(field_transformers=field_transformers)
    ht.fit(data)
    transformed = ht.transform(data)
    reverse_transformed = ht.reverse_transform(transformed)

    # Assert
    pd.testing.assert_frame_equal(transformed, expected_transformed)
    pd.testing.assert_frame_equal(expected_reversed, reverse_transformed)


def test_single_category():
    """Test that categorical variables with a single value are supported."""
    # Setup
    ht = HyperTransformer(field_transformers={
        'a': OneHotEncodingTransformer()
    })
    data = pd.DataFrame({
        'a': ['a', 'a', 'a']
    })

    # Run
    ht.fit(data)
    transformed = ht.transform(data)
    reverse = ht.reverse_transform(transformed)

    # Assert
    pd.testing.assert_frame_equal(data, reverse)


def test_dtype_category():
    """Test that categorical variables of dtype category are supported."""
    # Setup
    data = pd.DataFrame({'a': ['a', 'b', 'c']}, dtype='category')

    # Run
    ht = HyperTransformer()
    ht.fit(data)
    transformed = ht.transform(data)
    reverse = ht.reverse_transform(transformed)

    # Assert
    pd.testing.assert_frame_equal(reverse, data)


def test_subset_of_columns_nan_data():
    """HyperTransform should be able to transform a subset of the training columns.

    See https://github.com/sdv-dev/RDT/issues/152
    """
    # Setup
    data = get_input_data()
    subset = data[[data.columns[0]]].copy()
    ht = HyperTransformer()
    ht.fit(data)

    # Run
    transformed = ht.transform(subset)
    reverse = ht.reverse_transform(transformed)

    # Assert
    pd.testing.assert_frame_equal(subset, reverse)


def test_with_unfitted_columns():
    """HyperTransform should be able to transform even if there are unseen columns in data."""
    # Setup
    data = get_input_data()
    ht = HyperTransformer(default_data_type_transformers={'categorical': CategoricalTransformer})
    ht.fit(data)

    # Run
    new_data = get_input_data()
    new_column = pd.Series([4, 5, 6, 7, 8, 9])
    new_data['z'] = new_column
    transformed = ht.transform(new_data)
    reverse = ht.reverse_transform(transformed)

    # Assert
    expected_reversed = get_input_data()
    expected_reversed['z'] = new_column
    expected_reversed = expected_reversed.reindex(
        columns=['z', 'integer', 'float', 'categorical', 'bool', 'datetime', 'names'])
    pd.testing.assert_frame_equal(expected_reversed, reverse)



from unittest import TestCase
from unittest.mock import Mock, call, patch

import numpy as np
import pandas as pd
import pytest

from rdt import HyperTransformer
from rdt.errors import NotFittedError
from rdt.transformers import (
    BooleanTransformer, CategoricalTransformer, DatetimeTransformer, GaussianCopulaTransformer,
    NumericalTransformer, OneHotEncodingTransformer)


class TestHyperTransformer(TestCase):

    def test__validate_field_transformers(self):
        """Test the ``_validate_field_transformers`` method.

        Tests that a ``ValueError`` is raised if a field has multiple
        definitions in the ``field_transformers`` dict.

        Setup:
            - field_transformers set to have duplicate definitions.

        Expected behavior:
            - A ``ValueError`` should be raised if a field is defined
            more than once.
        """
        # Setup
        int_transformer = Mock()
        float_transformer = Mock()

        field_transformers = {
            'integer': int_transformer,
            'float': float_transformer,
            ('integer',): int_transformer
        }
        ht = HyperTransformer()
        ht.field_transformers = field_transformers

        # Run / Assert
        error_msg = (
            r'Multiple transformers specified for the field \(\'integer\',\). '
            'Each field can have at most one transformer defined in field_transformers.'
        )

        with pytest.raises(ValueError, match=error_msg):
            ht._validate_field_transformers()

    @patch('rdt.hyper_transformer.HyperTransformer._create_multi_column_fields')
    @patch('rdt.hyper_transformer.HyperTransformer._validate_field_transformers')
    def test___init__(self, validation_mock, multi_column_mock):
        """Test create new instance of HyperTransformer"""
        # Run
        ht = HyperTransformer()

        # Asserts
        assert ht.copy is True
        assert ht.field_transformers == {}
        assert ht.default_data_type_transformers == {}
        assert ht.field_data_types == {}
        multi_column_mock.assert_called_once()
        validation_mock.assert_called_once()

    def test__unfit(self):
        """Test the ``_unfit`` method.

        The ``_unfit`` method should reset ``instance._transformers_sequence``.
        It should also set ``_fitted`` to False.

        Setup:
            - instance._fitted is set to True
            - instance._transformers_sequence is a list of transformers

        Expected behavior:
            - instance._fitted is set to False
            - instance._transformers_sequence is set to []
        """
        # Setup
        ht = HyperTransformer()
        ht._transformers_sequence = [BooleanTransformer(), NumericalTransformer()]
        ht._fitted = True

        # Run
        ht._unfit()

        # Assert
        assert ht._fitted is False
        assert ht._transformers_sequence == []

    def test__create_multi_column_fields(self):
        """Test the ``_create_multi_column_fields`` method.

        This tests that the method goes through both the ``field_transformers``
        dict and ``field_data_types`` dict to find multi_column fields and map
        each column to its corresponding tuple.

        Setup:
            - instance.field_transformers will be populated with multi-column fields
            - instance.field_data_types will be populated with multi-column fields

        Output:
            - A dict mapping each column name that is part of a multi-column
            field to the tuple of columns in the field it belongs to.
        """
        # Setup
        ht = HyperTransformer()
        ht.field_transformers = {
            'a': BooleanTransformer,
            'b': DatetimeTransformer,
            ('c', 'd'): DatetimeTransformer,
            'e': NumericalTransformer
        }
        ht.field_data_types = {
            'f': 'categorical',
            ('g', 'h'): 'datetime'
        }

        # Run
        multi_column_fields = ht._create_multi_column_fields()

        # Assert
        expected = {
            'c': ('c', 'd'),
            'd': ('c', 'd'),
            'g': ('g', 'h'),
            'h': ('g', 'h')
        }
        assert multi_column_fields == expected

    def test__get_next_transformer_field_transformer(self):
        """Test the ``_get_next_transformer method.

        This tests that if the transformer is defined in the
        ``instance.field_transformers`` dict, then it is returned
        even if the output type is final.

        Setup:
            - field_transformers is given a transformer for the
            output field.
            - default_data_type_transformers will be given a different transformer
            for the output type of the output field.

        Input:
            - An output field name in field_transformers.
            - Output type is numerical.
            - next_transformers is None.

        Output:
            - The transformer defined in field_transformers.
        """
        # Setup
        transformer = NumericalTransformer()
        ht = HyperTransformer(
            field_transformers={'a.out': transformer},
            default_data_type_transformers={'numerical': GaussianCopulaTransformer()}
        )

        # Run
        next_transformer = ht._get_next_transformer('a.out', 'numerical', None)

        # Assert
        assert next_transformer == transformer

    def test__get_next_transformer_final_output_type(self):
        """Test the ``_get_next_transformer method.

        This tests that if the transformer is not defined in the
        ``instance.field_transformers`` dict and its output type
        is in ``instance._transform_output_types``, then ``None``
        is returned.

        Setup:
            - default_data_type_transformers will be given a transformer
            for the output type of the output field.

        Input:
            - An output field name in field_transformers.
            - Output type is numerical.
            - next_transformers is None.

        Output:
            - None.
        """
        # Setup
        ht = HyperTransformer(
            default_data_type_transformers={'numerical': GaussianCopulaTransformer()}
        )

        # Run
        next_transformer = ht._get_next_transformer('a.out', 'numerical', None)

        # Assert
        assert next_transformer is None

    def test__get_next_transformer_next_transformers(self):
        """Test the ``_get_next_transformer method.

        This tests that if the transformer is not defined in the
        ``instance.field_transformers`` dict and its output type
        is not in ``instance._transform_output_types`` and the
        ``next_transformers`` dict has a transformer for the output
        field, then it is used.

        Setup:
            - default_data_type_transformers will be given a transformer
            for the output type of the output field.

        Input:
            - An output field name in field_transformers.
            - Output type is categorical.
            - next_transformers is has a transformer defined
            for the output field.

        Output:
            - The transformer defined in next_transformers.
        """
        # Setup
        transformer = CategoricalTransformer()
        ht = HyperTransformer(
            default_data_type_transformers={'categorical': OneHotEncodingTransformer()}
        )
        next_transformers = {'a.out': transformer}

        # Run
        next_transformer = ht._get_next_transformer('a.out', 'categorical', next_transformers)

        # Assert
        assert next_transformer == transformer

    @patch('rdt.transformers.get_default_transformer')
    def test__get_next_transformer_default_transformer(self, mock):
        """Test the ``_get_next_transformer method.

        This tests that if the transformer is not defined in the
        ``instance.field_transformers`` dict or ``next_transformers``
        and its output typeis not in ``instance._transform_output_types``
        then the default_transformer is used.

        Setup:
            - A mock is used for ``get_default_transformer``.

        Input:
            - An output field name in field_transformers.
            - Output type is categorical.
            - next_transformers is None.

        Output:
            - The default transformer.
        """
        # Setup
        transformer = CategoricalTransformer(fuzzy=True)
        mock.return_value = transformer
        ht = HyperTransformer(
            default_data_type_transformers={'categorical': OneHotEncodingTransformer()}
        )

        # Run
        next_transformer = ht._get_next_transformer('a.out', 'categorical', None)

        # Assert
        assert isinstance(next_transformer, CategoricalTransformer)
        assert next_transformer.fuzzy is True

    def test__populate_field_data_types(self):
        """Test the ``_populate_field_data_types`` method.

        This tests that if any field types are missing in the
        provided field_data_types dict, that the rest of the values
        are filled in using the data types for the dtype.

        Setup:
            - field_data_types will only define a few of the fields.

        Input:
            - A DataFrame of various types.

        Expected behavior:
            - field types will have values for all fields in
            the data.
        """
        # Setup
        ht = HyperTransformer(field_data_types={'a': 'numerical', 'b': 'categorical'})
        data = pd.DataFrame({
            'a': [np.nan, 1, 2, 3],
            'b': [np.nan, 'category1', 'category2', 'category3'],
            'c': [np.nan, True, False, True],
            'd': [np.nan, 1.0, 2.0, 3.0]
        })

        # Run
        ht._populate_field_data_types(data)

        # Assert
        expected = {'a': 'numerical', 'b': 'categorical', 'c': 'boolean', 'd': 'float'}
        assert ht.field_data_types == expected

    @patch('rdt.hyper_transformer.get_transformer_instance')
    def test__fit_field_transformer(self, get_transformer_instance_mock):
        """Test the ``_fit_field_transformer`` method.

        This tests that the ``_fit_field_transformer`` behaves as expected.
        It should fit the transformer it is provided, loops through its
        outputs, transform the data if the output is not ML ready and call
        itself recursively if it can.

        Setup:
            - A mock for ``get_transformer_instance``.
            - A mock for the transformer returned by ``get_transformer_instance``.
            The ``get_output_types`` method will return two outputs, one that
            is ML ready and one that isn't.

        Input:
            - A DataFrame with one column.
            - A column name to fit the transformer to.
            - A transformer.

        Output:
            - A DataFrame with columns that result from transforming the
            outputs of the original transformer.
            - ``_output_columns`` should add the appropriate column names.
        """
        # Setup
        data = pd.DataFrame({'a': [1, 2, 3]})
        transformed_data1 = pd.DataFrame({
            'a.out1': ['2', '4', '6'],
            'a.out2': [1, 2, 3]
        })
        transformer1 = Mock()
        transformer2 = Mock()
        transformer1.get_output_types.return_value = {
            'a.out1': 'categorical',
            'a.out2': 'numerical'
        }
        transformer1.get_next_transformers.return_value = {
            'a.out1': transformer2
        }
        transformer1.transform.return_value = transformed_data1
        transformer2.get_output_types.return_value = {
            'a.out1.value': 'numerical'
        }
        transformer2.get_next_transformers.return_value = None
        get_transformer_instance_mock.side_effect = [
            transformer1,
            transformer2,
            None,
            None
        ]
        ht = HyperTransformer()
        ht._get_next_transformer = Mock()
        ht._get_next_transformer.side_effect = [
            transformer2,
            None,
            None
        ]

        # Run
        out = ht._fit_field_transformer(data, 'a', transformer1)

        # Assert
        expected = pd.DataFrame({
            'a.out1': ['2', '4', '6'],
            'a.out2': [1, 2, 3]
        })
        assert ht._output_columns == ['a.out1.value', 'a.out2']
        pd.testing.assert_frame_equal(out, expected)
        transformer1.fit.assert_called_once()
        transformer1.transform.assert_called_once_with(data)
        transformer2.fit.assert_called_once()
        assert ht._transformers_sequence == [transformer1, transformer2]
        assert ht._transformers_tree == {
            'a': {'transformer': transformer1, 'outputs': ['a.out1', 'a.out2']},
            'a.out1': {'transformer': transformer2, 'outputs': ['a.out1.value']}
        }

    @patch('rdt.hyper_transformer.get_transformer_instance')
    def test__fit_field_transformer_multi_column_field_not_ready(
        self,
        get_transformer_instance_mock
    ):
        """Test the ``_fit_field_transformer`` method.

        This tests that the ``_fit_field_transformer`` behaves as expected.
        If the column is part of a multi-column field, and the other columns
        aren't present in the data, then it should not fit the next transformer.
        It should however, transform the data.

        Setup:
            - A mock for ``get_transformer_instance``.
            - A mock for the transformer returned by ``get_transformer_instance``.
            The ``get_output_types`` method will return one output that is part of
            a multi-column field.
            - A mock for ``_multi_column_fields`` to return the multi-column field.

        Input:
            - A DataFrame with two columns.
            - A column name to fit the transformer to.
            - A transformer.

        Output:
            - A DataFrame with columns that result from transforming the
            outputs of the original transformer.
        """
        # Setup
        data = pd.DataFrame({
            'a': [1, 2, 3],
            'b': [4, 5, 6]
        })
        transformed_data1 = pd.DataFrame({
            'a.out1': ['1', '2', '3'],
            'b': [4, 5, 6]
        })
        transformer1 = Mock()
        transformer2 = Mock()
        transformer1.get_output_types.return_value = {
            'a.out1': 'categorical'
        }
        transformer1.get_next_transformers.return_value = None
        transformer1.transform.return_value = transformed_data1
        get_transformer_instance_mock.side_effect = [transformer1]
        ht = HyperTransformer()
        ht._get_next_transformer = Mock()
        ht._get_next_transformer.side_effect = [transformer2]
        ht._multi_column_fields = Mock()
        ht._multi_column_fields.get.return_value = ('a.out1', 'b.out1')

        # Run
        out = ht._fit_field_transformer(data, 'a', transformer1)

        # Assert
        expected = pd.DataFrame({
            'a.out1': ['1', '2', '3'],
            'b': [4, 5, 6]
        })
        assert ht._output_columns == []
        pd.testing.assert_frame_equal(out, expected)
        transformer1.fit.assert_called_once()
        transformer1.transform.assert_called_once_with(data)
        transformer2.fit.assert_not_called()
        assert ht._transformers_sequence == [transformer1]

    @patch('rdt.hyper_transformer.get_transformer_instance')
    def test__fit_field_transformer_multi_column_field_ready(self, get_transformer_instance_mock):
        """Test the ``_fit_field_transformer`` method.

        This tests that the ``_fit_field_transformer`` behaves as expected.
        If the column is part of a multi-column field, and the other columns
        are present in the data, then it should fit the next transformer.
        It should also transform the data.

        Setup:
            - A mock for ``get_transformer_instance``.
            - A mock for the transformer returned by ``get_transformer_instance``.
            The ``get_output_types`` method will return one output that is part of
            a multi-column field.
            - A mock for ``_multi_column_fields`` to return the multi-column field.

        Input:
            - A DataFrame with the other columns in the multi-column field.
            - A column name to fit the transformer to.
            - A transformer.

        Output:
            - A DataFrame with columns that result from transforming the
            outputs of the original transformer.
            - ``_output_columns`` should add the column name of the output of
            the transformer used on the multi-column field.
        """
        # Setup
        data = pd.DataFrame({
            'a': [1, 2, 3],
            'b.out1': ['4', '5', '6']
        })
        transformed_data1 = pd.DataFrame({
            'a.out1': ['1', '2', '3'],
            'b.out1': ['4', '5', '6']
        })
        transformer1 = Mock()
        transformer2 = Mock()
        transformer1.get_output_types.return_value = {
            'a.out1': 'categorical'
        }
        transformer1.get_next_transformers.return_value = None
        transformer1.transform.return_value = transformed_data1
        transformer2.get_output_types.return_value = {
            'a.out1#b.out1': 'numerical'
        }
        get_transformer_instance_mock.side_effect = [
            transformer1,
            transformer2
        ]
        ht = HyperTransformer()
        ht._get_next_transformer = Mock()
        ht._get_next_transformer.side_effect = [
            transformer2,
            None
        ]
        ht._multi_column_fields = Mock()
        ht._multi_column_fields.get.return_value = ('a.out1', 'b.out1')

        # Run
        out = ht._fit_field_transformer(data, 'a', transformer1)

        # Assert
        expected = pd.DataFrame({
            'a.out1': ['1', '2', '3'],
            'b.out1': ['4', '5', '6']
        })
        assert ht._output_columns == ['a.out1#b.out1']
        pd.testing.assert_frame_equal(out, expected)
        transformer1.fit.assert_called_once()
        transformer1.transform.assert_called_once_with(data)
        transformer2.fit.assert_called_once()
        transformer2.transform.assert_called_once()
        assert ht._transformers_sequence == [transformer1, transformer2]

    @patch('rdt.hyper_transformer.warnings')
    def test__validate_all_fields_fitted(self, warnings_mock):
        """Test the ``_validate_all_fields_fitted`` method.

        Tests that the ``_validate_all_fields_fitted`` method raises a warning
        if there are fields in ``field_transformers`` that were not fitted.

        Setup:
            - A mock for warnings.
            - A mock for ``_field_transformers`` with a misspelled field.
            - A mock for ``_fitted_fields`` containing the other fields.

        Expected behavior:
            - Warnings should be raised.
        """
        # Setup
        int_transformer = Mock()
        float_transformer = Mock()
        field_transformers = {
            'integer': int_transformer,
            'float': float_transformer,
            'intege': int_transformer
        }
        ht = HyperTransformer(field_transformers=field_transformers)
        ht._fitted_fields = {'integer', 'float'}

        # Run
        ht._validate_all_fields_fitted()

        # Assert
        warnings_mock.warn.assert_called_once()

    def get_data(self):
        return pd.DataFrame({
            'integer': [1, 2, 1, 3],
            'float': [0.1, 0.2, 0.1, 0.1],
            'categorical': ['a', 'a', 'b', 'a'],
            'bool': [False, False, True, False],
            'datetime': pd.to_datetime(['2010-02-01', '2010-01-01', '2010-02-01', '2010-01-01'])
        })

    def get_transformed_data(self, drop=False):
        data = pd.DataFrame({
            'integer': [1, 2, 1, 3],
            'float': [0.1, 0.2, 0.1, 0.1],
            'categorical': ['a', 'a', 'b', 'a'],
            'bool': [False, False, True, False],
            'datetime': pd.to_datetime(['2010-02-01', '2010-01-01', '2010-02-01', '2010-01-01']),
            'integer.out': ['1', '2', '1', '3'],
            'integer.out.value': [1, 2, 1, 3],
            'float.value': [0.1, 0.2, 0.1, 0.1],
            'categorical.value': [0.375, 0.375, 0.875, 0.375],
            'bool.value': [0.0, 0.0, 1.0, 0.0],
            'datetime.value': [
                1.2649824e+18,
                1.262304e+18,
                1.2649824e+18,
                1.262304e+18
            ]
        })

        if drop:
            return data.drop([
                'integer',
                'float',
                'categorical',
                'bool',
                'datetime',
                'integer.out'
            ], axis=1)

        return data

    @patch('rdt.hyper_transformer.get_default_transformer')
    def test_fit(self, get_default_transformer_mock):
        """Test the ``fit`` method.

        Tests that the ``fit`` method loops through the fields in ``field_transformers``
        and ``field_data_types`` that are in the data. It should try to find a transformer
        in ``default_data_type_transformers`` and then use the default if it doesn't find one
        when looping through ``field_data_types``. It should then call ``_fit_field_transformer``
        with the correct arguments.

        Setup:
            - A mock for ``_fit_field_transformer``.
            - A mock for ``_field_in_set``.
            - A mock for ``get_default_tranformer``.

        Input:
            - A DataFrame with multiple columns of different types.

        Expected behavior:
            - The ``_fit_field_transformer`` mock should be called with the correct
            arguments in the correct order.
        """
        # Setup
        int_transformer = Mock()
        int_out_transformer = Mock()
        float_transformer = Mock()
        categorical_transformer = Mock()
        bool_transformer = Mock()
        datetime_transformer = Mock()

        data = self.get_data()
        field_transformers = {
            'integer': int_transformer,
            'float': float_transformer,
            'integer.out': int_out_transformer
        }
        default_data_type_transformers = {
            'boolean': bool_transformer,
            'categorical': categorical_transformer
        }
        get_default_transformer_mock.return_value = datetime_transformer
        ht = HyperTransformer(
            field_transformers=field_transformers,
            default_data_type_transformers=default_data_type_transformers
        )
        ht._fit_field_transformer = Mock()
        ht._fit_field_transformer.return_value = data
        ht._field_in_set = Mock()
        ht._field_in_set.side_effect = [True, True, False, False, False]
        ht._validate_all_fields_fitted = Mock()

        # Run
        ht.fit(data)

        # Assert
        ht._fit_field_transformer.assert_has_calls([
            call(data, 'integer', int_transformer),
            call(data, 'float', float_transformer),
            call(data, 'categorical', categorical_transformer),
            call(data, 'bool', bool_transformer),
            call(data, 'datetime', datetime_transformer)
        ])
        ht._validate_all_fields_fitted.assert_called_once()

    def test_transform(self):
        """Test the ``transform`` method.

        Tests that ``transform`` loops through the ``_transformers_sequence``
        and calls ``transformer.transform`` in the correct order.

        Setup:
            - The ``_transformers_sequence`` will be hardcoded with a list
            of transformer mocks.
            - The ``_input_columns`` will be hardcoded.
            - The ``_output_columns`` will be hardcoded.

        Input:
            - A DataFrame of multiple types.

        Output:
            - The transformed DataFrame with the correct columns dropped.
        """
        # Setup
        int_transformer = Mock()
        int_out_transformer = Mock()
        float_transformer = Mock()
        categorical_transformer = Mock()
        bool_transformer = Mock()
        datetime_transformer = Mock()
        data = self.get_data()
        transformed_data = self.get_transformed_data()
        datetime_transformer.transform.return_value = transformed_data
        ht = HyperTransformer()
        ht._fitted = True
        ht._transformers_sequence = [
            int_transformer,
            int_out_transformer,
            float_transformer,
            categorical_transformer,
            bool_transformer,
            datetime_transformer
        ]
        ht._input_columns = list(data.columns)
        expected = self.get_transformed_data(True)
        ht._output_columns = list(expected.columns)

        # Run
        transformed = ht.transform(data)

        # Assert
        pd.testing.assert_frame_equal(transformed, expected)
        int_transformer.transform.assert_called_once()
        int_out_transformer.transform.assert_called_once()
        float_transformer.transform.assert_called_once()
        categorical_transformer.transform.assert_called_once()
        bool_transformer.transform.assert_called_once()
        datetime_transformer.transform.assert_called_once()

    def test_transform_raises_error_if_not_fitted(self):
        """Test that ``transform`` raises an error.

        The ``transform`` method should raise a ``NotFittedError`` if the
        ``_transformers_sequence`` is empty.

        Setup:
            - The ``_fitted`` attribute will be False.

        Input:
            - A DataFrame of multiple types.

        Expected behavior:
            - A ``NotFittedError`` is raised.
        """
        # Setup
        data = self.get_data()
        ht = HyperTransformer()

        # Run
        with pytest.raises(NotFittedError):
            ht.transform(data)

    def test_fit_transform(self):
        """Test call fit_transform"""
        # Run
        transformer = Mock()

        HyperTransformer.fit_transform(transformer, pd.DataFrame())

        # Asserts
        expect_call_count_fit = 1
        expect_call_count_transform = 1
        expect_call_args_fit = pd.DataFrame()
        expect_call_args_transform = pd.DataFrame()

        assert transformer.fit.call_count == expect_call_count_fit
        pd.testing.assert_frame_equal(
            transformer.fit.call_args[0][0],
            expect_call_args_fit
        )

        assert transformer.transform.call_count == expect_call_count_transform
        pd.testing.assert_frame_equal(
            transformer.transform.call_args[0][0],
            expect_call_args_transform
        )

    def test_reverse_transform(self):
        """Test the ``reverse_transform`` method.

        Tests that ``reverse_transform`` loops through the ``_transformers_sequence``
        in reverse order and calls ``transformer.reverse_transform``.

        Setup:
            - The ``_transformers_sequence`` will be hardcoded with a list
            of transformer mocks.
            - The ``_output_columns`` will be hardcoded.
            - The ``_input_columns`` will be hardcoded.

        Input:
            - A DataFrame of multiple types.

        Output:
            - The reverse transformed DataFrame with the correct columns dropped.
        """
        # Setup
        int_transformer = Mock()
        int_out_transformer = Mock()
        float_transformer = Mock()
        categorical_transformer = Mock()
        bool_transformer = Mock()
        datetime_transformer = Mock()
        data = self.get_transformed_data(True)
        reverse_transformed_data = self.get_transformed_data()
        int_transformer.reverse_transform.return_value = reverse_transformed_data
        ht = HyperTransformer()
        ht._fitted = True
        ht._transformers_sequence = [
            int_transformer,
            int_out_transformer,
            float_transformer,
            categorical_transformer,
            bool_transformer,
            datetime_transformer
        ]
        ht._output_columns = list(data.columns)
        expected = self.get_data()
        ht._input_columns = list(expected.columns)

        # Run
        reverse_transformed = ht.reverse_transform(data)

        # Assert
        pd.testing.assert_frame_equal(reverse_transformed, expected)
        int_transformer.reverse_transform.assert_called_once()
        int_out_transformer.reverse_transform.assert_called_once()
        float_transformer.reverse_transform.assert_called_once()
        categorical_transformer.reverse_transform.assert_called_once()
        bool_transformer.reverse_transform.assert_called_once()
        datetime_transformer.reverse_transform.assert_called_once()

    def test_reverse_transform_raises_error_if_not_fitted(self):
        """Test that ``reverse_transform`` raises an error.

        The ``reverse_transform`` method should raise a ``NotFittedError`` if the
        ``_transformers_sequence`` is empty.

        Setup:
            - The ``_fitted`` attribute will be False.

        Input:
            - A DataFrame of multiple types.

        Expected behavior:
            - A ``NotFittedError`` is raised.
        """
        # Setup
        data = self.get_transformed_data()
        ht = HyperTransformer()

        # Run
        with pytest.raises(NotFittedError):
            ht.reverse_transform(data)

    def test_get_field_data_types(self):
        """Test the ``get_field_data_types`` method.

        This method should return the ``field_data_types`` attribute.

        Output:
            - Dict mapping fields to data types.
        """
        # Setup
        field_data_types = {
            'a': 'categorical',
            'b': 'integer'
        }
        ht = HyperTransformer(field_data_types=field_data_types)

        # Run
        out = ht.get_field_data_types()

        # Assert
        assert out == {'a': 'categorical', 'b': 'integer'}

    def test_update_field_data_types(self):
        """Test the ``update_field_data_types`` method.

        This method should update the ``field_data_types`` attribute.

        Setup:
            - Initialize ``HyperTransformer`` with ``field_data_types`` having
            one entry.

        Input:
            - Dict mapping fields to data types.
        """
        # Setup
        field_data_types = {
            'a': 'categorical',
            'b': 'integer'
        }
        ht = HyperTransformer(field_data_types={'a': 'float'})
        ht._transformers_sequence = [CategoricalTransformer()]
        ht._unfit = Mock()

        # Run
        ht.update_field_data_types(field_data_types)

        # Assert
        assert ht.field_data_types == {'a': 'categorical', 'b': 'integer'}
        ht._unfit.assert_called_once()

    def test_get_default_data_type_transformers(self):
        """Test the ``get_default_data_type_transformers`` method.

        This method should return the ``default_data_type_transformers`` attribute.

        Output:
            - Dict mapping data types to transformers.
        """
        # Setup
        data_type_transformers = {
            'categorical': CategoricalTransformer,
            'integer': NumericalTransformer
        }
        ht = HyperTransformer(default_data_type_transformers=data_type_transformers)

        # Run
        out = ht.get_default_data_type_transformers()

        # Assert
        assert out == {'categorical': CategoricalTransformer, 'integer': NumericalTransformer}

    def test_update_default_data_type_transformers(self):
        """Test the ``update_default_data_type_transformers`` method.

        This method should update the ``default_data_type_transformers`` attribute.

        Setup:
            - Initialize ``HyperTransformer`` with ``default_data_type_transformers``
            dict that only has some types set.

        Input:
            - Dict mapping new data types to transformers.
        """
        # Setup
        data_type_transformers = {
            'categorical': CategoricalTransformer,
            'integer': NumericalTransformer
        }
        ht = HyperTransformer(default_data_type_transformers=data_type_transformers)
        ht._transformers_sequence = [CategoricalTransformer()]
        ht._unfit = Mock()

        # Run
        ht.update_default_data_type_transformers({'boolean': BooleanTransformer})

        # Assert
        assert ht.default_data_type_transformers == {
            'categorical': CategoricalTransformer,
            'integer': NumericalTransformer,
            'boolean': BooleanTransformer
        }
        ht._unfit.assert_called_once()

    def test_set_first_transformers_for_fields(self):
        """Test the ``set_first_transformers_for_fields`` method.

        This method should update the ``field_transformers`` attribute.

        Setup:
            - Initialize ``HyperTransformer`` with ``field_transformers`` dict that only
            has some fields set.

        Input:
            - Dict mapping one new field to a transformer and one old field to a different
            transformer.
        """
        # Setup
        field_transformers = {
            'a': CategoricalTransformer,
            'b': NumericalTransformer
        }
        ht = HyperTransformer(field_transformers=field_transformers)
        ht._transformers_sequence = [CategoricalTransformer()]
        ht._unfit = Mock()

        # Run
        ht.set_first_transformers_for_fields({
            'c': BooleanTransformer,
            'b': CategoricalTransformer
        })

        # Assert
        assert ht.field_transformers == {
            'a': CategoricalTransformer,
            'b': CategoricalTransformer,
            'c': BooleanTransformer
        }
        ht._unfit.assert_called_once()

    def test_get_transformer(self):
        """Test the ``get_transformer`` method.

        The method should return the transformer instance stored for the field in the
        ``_transformers_tree``.

        Setup:
            - Set the ``_transformers_tree`` to have values for a few fields.
            - Set ``_fitted`` to ``True``.

        Input:
            - Each field name.

        Output:
            - The transformer instance used for each field.
        """
        # Setup
        ht = HyperTransformer()
        transformer1 = Mock()
        transformer2 = Mock()
        transformer3 = Mock()
        ht._transformers_tree = {
            'field': {'transformer': transformer1, 'outputs': ['field.out1', 'field.out2']},
            'field.out1': {'transformer': transformer2, 'outputs': ['field.out1.value']},
            'field.out2': {'transformer': transformer3, 'outputs': ['field.out2.value']}
        }
        ht._fitted = True

        # Run
        out1 = ht.get_transformer('field')
        out2 = ht.get_transformer('field.out1')
        out3 = ht.get_transformer('field.out2')

        # Assert
        out1 == transformer1
        out2 == transformer2
        out3 == transformer3

    def test_get_transformer_raises_error_if_not_fitted(self):
        """Test that ``get_transformer`` raises ``NotFittedError``.

        If the ``HyperTransformer`` hasn't been fitted, a ``NotFittedError`` should be raised.

        Setup:
            - Set ``_fitted`` to ``False``.

        Expected behavior:
            - ``NotFittedError`` is raised.
        """
        # Setup
        ht = HyperTransformer()
        transformer1 = Mock()
        ht._transformers_tree = {
            'field1': {'transformer': transformer1, 'outputs': ['field1.out1', 'field1.out2']}
        }
        ht._fitted = False

        # Run / Assert
        with pytest.raises(NotFittedError):
            ht.get_transformer('field1')

    def test_get_output_transformers(self):
        """Test the ``get_output_transformers`` method.

        The method should return a dict mapping each output column created from
        transforming the specified field, to the transformers to be used on them.

        Setup:
            - Set the ``_transformers_tree`` to have values for a few fields.
            - Set ``_fitted`` to ``False``.

        Output:
            - Dict mapping the outputs of the specified field to the transformers used on them.
        """
        # Setup
        ht = HyperTransformer()
        transformer1 = Mock()
        transformer2 = Mock()
        transformer3 = Mock()
        transformer4 = Mock()
        ht._transformers_tree = {
            'field1': {'transformer': transformer1, 'outputs': ['field1.out1', 'field1.out2']},
            'field1.out1': {'transformer': transformer2, 'outputs': ['field1.out1.value']},
            'field1.out2': {'transformer': transformer3, 'outputs': ['field1.out2.value']},
            'field2': {'transformer': transformer4, 'outputs': ['field2.value']}
        }
        ht._fitted = True

        # Run
        output_transformers = ht.get_output_transformers('field1')

        # Assert
        output_transformers == {
            'field1.out1': transformer2,
            'field1.out2': transformer3
        }

    def test_get_output_transformers_raises_error_if_not_fitted(self):
        """Test that the ``get_output_transformers`` raises ``NotFittedError``.

        If the ``HyperTransformer`` hasn't been fitted, a ``NotFittedError`` should be raised.

        Setup:
            - Set ``_fitted`` to ``False``.

        Expected behavior:
            - ``NotFittedError`` is raised.
        """
        # Setup
        ht = HyperTransformer()
        ht._fitted = False

        # Run / Assert
        with pytest.raises(NotFittedError):
            ht.get_output_transformers('field')

    def test_get_final_output_columns(self):
        """Test the ``get_fianl_output_columns`` method.

        The method should traverse the tree starting at the provided field and return a list of
        all final column names (leaf nodes) that are descedants of that field.

        Setup:
            - Set the ``_transformers_tree`` to have some fields with multiple generations
            of descendants.
            - Set ``_fitted`` to ``False``.

        Output:
            - List of all column nodes with the specified field as an ancestor.
        """
        # Setup
        ht = HyperTransformer()
        transformer = Mock()
        ht._transformers_tree = {
            'field1': {'transformer': transformer, 'outputs': ['field1.out1', 'field1.out2']},
            'field1.out1': {'transformer': transformer, 'outputs': ['field1.out1.value']},
            'field1.out2': {
                'transformer': transformer,
                'outputs': ['field1.out2.out1', 'field.out2.out2']
            },
            'field1.out2.out1': {
                'transformer': transformer,
                'outputs': ['field1.out2.out1.value']
            },
            'field1.out2.out2': {
                'transformer': transformer,
                'outputs': ['field1.out2.out2.value']
            },
            'field2': {'transformer': transformer, 'outputs': ['field2.value']}
        }
        ht._fitted = True

        # Run
        outputs = ht.get_final_output_columns('field1')

        # Assert
        outputs == ['field1.out1.value', 'field1.out2.out1.value', 'field1.out2.out2.value']

    def test_get_final_output_columns_raises_error_if_not_fitted(self):
        """Test that the ``get_final_output_columns`` raises ``NotFittedError``.

        If the ``HyperTransformer`` hasn't been fitted, a ``NotFittedError`` should be raised.

        Setup:
            - Set ``_fitted`` to ``False``.

        Expected behavior:
            - ``NotFittedError`` is raised.
        """
        # Setup
        ht = HyperTransformer()
        ht._fitted = False

        # Run / Assert
        with pytest.raises(NotFittedError):
            ht.get_final_output_columns('field')

    def test_get_transformer_tree_yaml(self):
        """Test the ``get_transformer_tree_yaml`` method.

        The method should return a YAML representation of the tree.

        Setup:
            - Set the ``_transformers_tree`` to have values for a few fields.

        Output:
            - YAML object rrepresenting tree.
        """
        # Setup
        ht = HyperTransformer()
        ht._transformers_tree = {
            'field1': {
                'transformer': CategoricalTransformer(),
                'outputs': ['field1.out1', 'field1.out2']
            },
            'field1.out1': {
                'transformer': NumericalTransformer(),
                'outputs': ['field1.out1.value']
            },
            'field1.out2': {
                'transformer': NumericalTransformer(),
                'outputs': ['field1.out2.value']
            },
            'field2': {'transformer': CategoricalTransformer(), 'outputs': ['field2.value']}
        }
        ht._fitted = True

        # Run
        tree_yaml = ht.get_transformer_tree_yaml()

        # Assert
        assert tree_yaml == (
            'field1:\n  outputs:\n  - field1.out1\n  - field1.out2\n  '
            'transformer: CategoricalTransformer\nfield1.out1:\n  outputs:\n  - '
            'field1.out1.value\n  transformer: NumericalTransformer\nfield1.out2:\n  '
            'outputs:\n  - field1.out2.value\n  transformer: NumericalTransformer\nfield2:\n  '
            'outputs:\n  - field2.value\n  transformer: CategoricalTransformer\n'
        )



import numpy as np
import pandas as pd

from rdt.transformers import BooleanTransformer


class TestBooleanTransformer:

    def test_boolean_some_nans(self):
        """Test BooleanTransformer on input with some nan values.

        Ensure that the BooleanTransformer can fit, transform, and reverse
        transform on boolean data with Nones. Expect that the reverse
        transformed data is the same as the input.

        Input:
            - boolean data with None values
        Output:
            - The reversed transformed data
        """
        # Setup
        data = pd.DataFrame([True, False, None, False], columns=['bool'])
        transformer = BooleanTransformer()

        # Run
        transformer.fit(data, data.columns.to_list())
        transformed = transformer.transform(data)
        reverse = transformer.reverse_transform(transformed)

        # Assert
        pd.testing.assert_frame_equal(reverse, data)

    def test_boolean_all_nans(self):
        """Test BooleanTransformer on input with all nan values.

        Ensure that the BooleanTransformer can fit, transform, and reverse
        transform on boolean data with all Nones. Expect that the reverse
        transformed data is the same as the input.

        Input:
            - 4 rows of all None values
        Output:
            - The reversed transformed data
        """
        # Setup
        data = pd.DataFrame([None, None, None, None], columns=['bool'])
        transformer = BooleanTransformer()

        # Run
        transformer.fit(data, data.columns.to_list())
        transformed = transformer.transform(data)
        reverse = transformer.reverse_transform(transformed)

        # Assert
        pd.testing.assert_frame_equal(reverse, data)

    def test_boolean_input_unchanged(self):
        """Test BooleanTransformer on input with some nan values.

        Ensure that the BooleanTransformer can fit, transform, and reverse
        transform on boolean data with all Nones. Expect that the intermediate
        transformed data is unchanged.

        Input:
            - 4 rows of all None values
        Output:
            - The reversed transformed data
        Side effects:
            - The intermediate transformed data is unchanged.
        """
        # Setup
        data = pd.DataFrame([True, False, None, False], columns=['bool'])
        transformer = BooleanTransformer()

        # Run
        transformer.fit(data, data.columns.to_list())
        transformed = transformer.transform(data)
        unchanged_transformed = transformed.copy()
        reverse = transformer.reverse_transform(transformed)

        # Assert
        pd.testing.assert_frame_equal(reverse, data)
        np.testing.assert_array_equal(unchanged_transformed, transformed)



from rdt import get_demo


def test_get_demo():
    demo = get_demo()

    assert list(demo.columns) == ['0_int', '1_float', '2_str', '3_datetime']
    assert len(demo) == 10
    assert list(demo.isna().sum(axis=0)) == [2, 2, 2, 2]



"""Unit tests for the NullTransformer."""

import numpy as np
import pandas as pd
import pytest

from rdt.transformers import NullTransformer


class TestNullTransformer:

    def test___init__default(self):
        """Test the initialization without passing any default arguments.

        When no arguments are passed, the attributes should be populated
        with the right values.

        Input:
            - nothing

        Expected Side Effects:
            - The `fill_value` attribute should be `None`.
            - The `null_column` attribute should be `None`.
            - The `copy` attribute should be `False`.
        """
        # Run
        transformer = NullTransformer()

        # Assert
        assert transformer.fill_value is None
        assert transformer.null_column is None
        assert not transformer.copy

    def test___init__not_default(self):
        """Test the initialization passing values different than defaults.

        When arguments are passed, the attributes should be populated
        with the right values.

        Input:
            - Values different than the defaults.

        Expected Side Effects:
            - The attributes should be populated with the given values.
        """
        # Run
        transformer = NullTransformer('a_fill_value', False, True)

        # Assert
        assert transformer.fill_value == 'a_fill_value'
        assert not transformer.null_column
        assert transformer.copy

    def test_creates_null_column(self):
        """Test the creates_null_column method.

        If the `null_column` attributes evalutes to True, the `create_null_column`
        method should return the same value.

        Setup:
            - Create an instance and set _null_column to True

        Expected Output:
            - True
        """
        # Setup
        transformer = NullTransformer('something', null_column=True)
        transformer._null_column = True

        # Run
        creates_null_column = transformer.creates_null_column()

        # Assert
        assert creates_null_column

    def test__get_fill_value_scalar(self):
        """Test _get_fill_value when a scalar value is passed.

        If a fill_value different from None, 'mean' or 'mode' is
        passed to __init__, that value is returned.

        Setup:
            - NullTransformer passing a specific fill_value
              that is not None, mean or mode.

        Input:
            - A Series with some values.
            - A np.array with boolean values.

        Expected Output:
            - The value passed to __init__
        """
        # Setup
        transformer = NullTransformer('a_fill_value')

        # Run
        data = pd.Series([1, np.nan, 3], name='abc')
        null_values = np.array([False, True, False])
        fill_value = transformer._get_fill_value(data, null_values)

        # Assert
        assert fill_value == 'a_fill_value'

    def test__get_fill_value_all_nulls(self):
        """Test _get_fill_value when all the values are null.

        If the fill_value is not a scalar value and all the data
        values are null, the output should be 0.

        Setup:
            - NullTransformer passing 'mean' as the fill_value.

        Input:
            - A Series filled with nan values.
            - A np.array of all True values.

        Expected Output:
            - 0
        """
        # Setup
        transformer = NullTransformer('mean')

        # Run
        data = pd.Series([np.nan, np.nan, np.nan], name='abc')
        null_values = np.array([True, True, True])
        fill_value = transformer._get_fill_value(data, null_values)

        # Assert
        assert fill_value == 0

    def test__get_fill_value_none_numerical(self):
        """Test _get_fill_value when fill_value is None and data is numerical.

        If the fill_value is None and the data is numerical,
        the output fill value should be the mean of the input data.

        Setup:
            - NullTransformer passing with default arguments.

        Input:
            - An Series filled with integer values such that the mean
              is not contained in the series and there is at least one null.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The mean of the inputted Series.
        """
        # Setup
        transformer = NullTransformer()

        # Run
        data = pd.Series([1, 2, np.nan], name='abc')
        null_values = np.array([False, False, True])
        fill_value = transformer._get_fill_value(data, null_values)

        # Assert
        assert fill_value == 1.5

    def test__get_fill_value_none_not_numerical(self):
        """Test _get_fill_value when fill_value is None and data is not numerical.

        If the fill_value is None and the data is not numerical,
        the output fill value should be the mode of the input data.

        Setup:
            - NullTransformer with default arguments.

        Input:
            - An Series filled with string values with variable frequency and
              at least one null value.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The most frequent value in the input series.
        """
        # Setup
        transformer = NullTransformer()

        # Run
        data = pd.Series(['a', 'b', 'b', np.nan], name='abc')
        null_values = np.array([False, False, False, True])
        fill_value = transformer._get_fill_value(data, null_values)

        # Assert
        assert fill_value == 'b'

    def test__get_fill_value_mean(self):
        """Test _get_fill_value when fill_value is mean.

        If the fill_value is mean the output fill value should be the
        mean of the input data.

        Setup:
            - NullTransformer passing 'mean' as the fill_value.

        Input:
            - A Series filled with integer values such that the mean
              is not contained in the series and there is at least one null.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The mode of the inputted Series.
        """
        # Setup
        transformer = NullTransformer('mean')

        # Run
        data = pd.Series([1, 2, np.nan], name='abc')
        null_values = np.array([False, False, True])
        fill_value = transformer._get_fill_value(data, null_values)

        # Assert
        assert fill_value == 1.5

    def test__get_fill_value_mode(self):
        """Test _get_fill_value when fill_value is 'mode'.

        If the fill_value is 'mode' the output fill value should be the
        mode of the input data.

        Setup:
            - NullTransformer passing 'mode' as the fill_value.

        Input:
            - A Series filled with integer values such that the mean
              is not contained in the series and there is at least one null.
            - A np.array of booleans indicating which values are null.

        Expected Output:
            - The most frequent value in the input series.
        """
        # Setup
        transformer = NullTransformer('mode')

        # Run
        data = pd.Series([1, 2, 2, np.nan], name='abc')
        null_values = np.array([False, False, False, True])
        fill_value = transformer._get_fill_value(data, null_values)

        # Assert
        assert fill_value == 2

    def test_fit_null_column_none_and_nulls(self):
        """Test fit when null column is none and there are nulls.

        If there are nulls in the data and null_column was given as None,
        then the _null_column attribute should be set to True.
        Also validate that the null attribute and the _fill_value attributes
        are set accordingly.

        Setup:
            - A NullTransformer with default arguments.

        Input:
            - pd.Series of integers that contains nulls.

        Expected Side Effects:
            - the null_column attribute should be set to True.
            - the nulls attribute should be set to True.
            - the fill_value should be set to the mean of the given integers.
        """
        # Setup
        transformer = NullTransformer()

        # Run
        data = pd.Series([1, 2, np.nan])
        transformer.fit(data)

        # Assert
        assert transformer.nulls
        assert transformer._null_column
        assert transformer._fill_value == 1.5

    def test_fit_null_column_none_and_no_nulls(self):
        """Test fit when null column is none and there are NO nulls.

        If there are no nulls in the data and null_column was given as None,
        then the _null_column attribute should be set to False.
        Also validate that the null attribute and the _fill_value attributes
        are set accordingly.

        Setup:
            - A NullTransformer with default arguments.

        Input:
            - pd.Series of strings that contains no nulls.

        Expected Side Effects:
            - the null_column attribute should be set to False.
            - the nulls attribute should be set to False.
            - the fill_value should be set to the mode of the given strings.
        """
        # Setup
        transformer = NullTransformer()

        # Run
        data = pd.Series(['a', 'b', 'b'])
        transformer.fit(data)

        # Assert
        assert not transformer.nulls
        assert not transformer._null_column
        assert transformer._fill_value == 'b'

    def test_fit_null_column_not_none(self):
        """Test fit when null column is set to True/False.

        If null_column is set to True or False, the _null_column should
        get that value regardless of whether there are nulls or not.

        Notice that this test covers 4 scenarios at once.

        Setup:
            - 4 NullTransformer intances, 2 of them passing False for the null_column
              and 2 of them passing True.

        Input:
            - 2 pd.Series, one containing nulls and the other not containing nulls.

        Expected Side Effects:
            - the _null_column attribute should be set to True or False as indicated
              in the Transformer creation.
            - the nulls attribute should be True or False depending on whether
              the input data contains nulls or not.
        """
        # Setup
        null_column_false_nulls = NullTransformer(null_column=False)
        null_column_false_no_nulls = NullTransformer(null_column=False)
        null_column_true_nulls = NullTransformer(null_column=True)
        null_column_true_no_nulls = NullTransformer(null_column=True)
        nulls_str = pd.Series(['a', 'b', 'b', np.nan])
        no_nulls_str = pd.Series(['a', 'b', 'b', 'c'])
        nulls_int = pd.Series([1, 2, 3, np.nan])
        no_nulls_int = pd.Series([1, 2, 3, 4])

        # Run
        null_column_false_nulls.fit(nulls_str)
        null_column_false_no_nulls.fit(no_nulls_str)
        null_column_true_nulls.fit(nulls_int)
        null_column_true_no_nulls.fit(no_nulls_int)

        # Assert
        assert not null_column_false_nulls._null_column
        assert null_column_false_nulls.nulls
        assert null_column_false_nulls._fill_value == 'b'

        assert not null_column_false_no_nulls._null_column
        assert not null_column_false_no_nulls.nulls
        assert null_column_false_no_nulls._fill_value == 'b'

        assert null_column_true_nulls._null_column
        assert null_column_true_nulls.nulls
        assert null_column_true_nulls._fill_value == 2

        assert null_column_true_no_nulls._null_column
        assert not null_column_true_no_nulls.nulls
        assert null_column_true_no_nulls._fill_value == 2.5

    def test_transform__null_column_false_copy_false(self):
        """Test transform when _null_column is False and copy is False.

        When the _null_column is false, the nulls should be replaced
        by the _fill_value.
        When copy is False, the original data is affected.

        Setup:
            - NullTransformer instance with _null_column set to False,
              _fill_value set to a scalar value, and copy set to False.

        Input:
            - A pd.Series of integers with nulls.

        Expected Output:
            - Same data as the input, replacing the nulls with the
              scalar value.

        Expected Side Effects:
            - The input data has the null values replaced.
        """
        # Setup
        transformer = NullTransformer()
        transformer._null_column = False
        transformer._fill_value = 3
        transformer.copy = False
        input_data = pd.Series([1, 2, np.nan])

        # Run
        output = transformer.transform(input_data)

        # Assert
        expected_output = np.array([1, 2, 3])
        np.testing.assert_equal(expected_output, output)

        modified_input_data = pd.Series([1, 2, 3], dtype=np.float64)
        pd.testing.assert_series_equal(modified_input_data, input_data)

    def test_transform__null_column_true_copy_false(self):
        """Test transform when _null_column and copy is False.

        When _null_column, the nulls should be replaced
        by the _fill_value and another column flagging the nulls
        should be created.
        When copy is False, the original data is affected.

        Setup:
            - NullTransformer instance with _null_column set to True,
              _fill_value set to a scalar value, and copy set to False.

        Input:
            - A pd.Series of strings with nulls.

        Expected Output:
            - Exactly the same as the input, replacing the nulls with the
              scalar value.

        Expected Side Effects:
            - The input data has the null values replaced.
        """
        # Setup
        transformer = NullTransformer()
        transformer.nulls = False
        transformer._null_column = True
        transformer._fill_value = 'c'
        transformer.copy = False
        input_data = pd.Series(['a', 'b', np.nan])

        # Run
        output = transformer.transform(input_data)

        # Assert
        expected_output = np.array([
            ['a', 0.0],
            ['b', 0.0],
            ['c', 1.0],
        ], dtype=object)
        np.testing.assert_equal(expected_output, output)

        modified_input_data = pd.Series(['a', 'b', 'c'])
        pd.testing.assert_series_equal(modified_input_data, input_data)

    def test_transform__null_column_false_copy_true(self):
        """Test transform when _null_column is False and copy.

        When the _null_column is false, the nulls should be replaced
        by the _fill_value.
        When copy, the original data is not affected.

        Setup:
            - NullTransformer instance with _null_column set to False,
              _fill_value set to a scalar value, and copy set to True.

        Input:
            - A pd.Series of integers with nulls.

        Expected Output:
            - Same data as the input, replacing the nulls with the
              scalar value.

        Expected Side Effects:
            - The input data has not been modified.
        """
        # Setup
        transformer = NullTransformer()
        transformer._null_column = False
        transformer._fill_value = 3
        transformer.copy = True
        input_data = pd.Series([1, 2, np.nan])

        # Run
        output = transformer.transform(input_data)

        # Assert
        expected_output = np.array([1, 2, 3])
        np.testing.assert_equal(expected_output, output)

        modified_input_data = pd.Series([1, 2, np.nan])
        pd.testing.assert_series_equal(modified_input_data, input_data)

    def test_transform__null_column_true_copy_true(self):
        """Test transform when _null_column and copy.

        When _null_column, the nulls should be replaced
        by the _fill_value and another column flagging the nulls
        should be created.
        When copy, the original data is not affected.

        Setup:
            - NullTransformer instance with _null_column set to True,
              _fill_value set to a scalar value, and copy set to True.

        Input:
            - A pd.Series of strings with nulls.

        Expected Output:
            - Exactly the same as the input, replacing the nulls with the
              scalar value.

        Expected Side Effects:
            - The input data has not been modified.
        """
        # Setup
        transformer = NullTransformer()
        transformer.nulls = False
        transformer._null_column = True
        transformer._fill_value = 'c'
        transformer.copy = True
        input_data = pd.Series(['a', 'b', np.nan])

        # Run
        output = transformer.transform(input_data)

        # Assert
        expected_output = np.array([
            ['a', 0.0],
            ['b', 0.0],
            ['c', 1.0],
        ], dtype=object)
        np.testing.assert_equal(expected_output, output)

        modified_input_data = pd.Series(['a', 'b', np.nan])
        pd.testing.assert_series_equal(modified_input_data, input_data)

    def test_transform__irreversible_warning(self):
        """Test transform when _null_column is False and fill_value is in data.

        When _null_column is False, and the fill_value exists in the data
        before transforming, a warning is raised.

        Setup:
            - NullTransformer instance with _null_column set to False
              and _fill_value set to a scalar value.

        Input:
            - A pd.Series of strings with nulls, containing the fill_value.

        Expected Output:
            - Exactly the same as the input, replacing the nulls with the
              scalar value.

        Expected Side Effects:
            - A UserWarning is raised.
        """
        # Setup
        transformer = NullTransformer()
        transformer._null_column = False
        transformer._fill_value = 'b'
        input_data = pd.Series(['a', 'b', np.nan])

        # Run
        with pytest.warns(UserWarning):
            output = transformer.transform(input_data)

        # Assert
        expected_output = np.array(['a', 'b', 'b'])
        np.testing.assert_equal(expected_output, output)

    def test_reverse_transform__null_column_true_nulls_true_copy_true(self):
        """Test reverse_transform when _null_column, nulls and copy are True.

        When _null_column and nulls attributes are both True, the second column
        in the input data should be used to decide which values to replace
        with nan, by selecting the rows where the null column value is > 0.5.
        If copy, the input data should not be modified.

        Setup:
            - NullTransformer instance with _null_column, nulls and copy
              attributes set to True.

        Input:
            - 2d numpy array with variate float values.

        Expected Output:
            - pd.Series containing the first column from the input data
              with the values indicated by the first column replaced by nans.

        Expected Side Effects:
            - the input data should not have been modified.
        """
        # Setup
        transformer = NullTransformer()
        transformer._null_column = True
        transformer.nulls = True
        transformer.copy = True
        input_data = np.array([
            [0.0, 0.0],
            [0.2, 0.2],
            [0.4, 0.4],
            [0.6, 0.6],
            [0.8, 0.8],
        ])
        input_data_copy = input_data.copy()

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, 0.4, np.nan, np.nan])
        pd.testing.assert_series_equal(expected_output, output)

        np.testing.assert_equal(input_data, input_data_copy)

    def test_reverse_transform__null_column_true_nulls_true_copy_false(self):
        """Test reverse_transform when _null_column and nulls are True, and copy is False.

        When _null_column and nulls attributes are both True, the second column
        in the input data should be used to decide which values to replace
        with nan, by selecting the rows where the null column value is > 0.5.
        If copy is False, the input data should be modified in place.

        Setup:
            - NullTransformer instance with _null_column and nulls
              attributes set to True, and copy set to False.

        Input:
            - 2d numpy array with variate float values.

        Expected Output:
            - pd.Series containing the first column from the input data
              with the values indicated by the first column replaced by nans.

        Expected Side Effects:
            - the input data should have been modified.
        """
        # Setup
        transformer = NullTransformer()
        transformer._null_column = True
        transformer.nulls = True
        transformer.copy = False
        input_data = np.array([
            [0.0, 0.0],
            [0.2, 0.2],
            [0.4, 0.4],
            [0.6, 0.6],
            [0.8, 0.8],
        ])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, 0.4, np.nan, np.nan])
        pd.testing.assert_series_equal(expected_output, output)

        modified_input_data = np.array([
            [0.0, 0.0],
            [0.2, 0.2],
            [0.4, 0.4],
            [np.nan, 0.6],
            [np.nan, 0.8],
        ])
        np.testing.assert_equal(modified_input_data, input_data)

    def test_reverse_transform__null_column_true_nulls_false(self):
        """Test reverse_transform when _null_column and nulls is False.

        When _null_column but nulls are False, the second column of the
        input data must be dropped and the first one returned as a Series without
        having been modified.

        Setup:
            - NullTransformer instance with _null_column set to True and nulls
              attribute set to False

        Input:
            - 2d numpy array with variate float values.

        Expected Output:
            - pd.Series containing the first column from the input data unmodified.
        """
        # Setup
        transformer = NullTransformer()
        transformer._null_column = True
        transformer.nulls = False
        input_data = np.array([
            [0.0, 0.0],
            [0.2, 0.2],
            [0.4, 0.4],
            [0.6, 0.6],
            [0.8, 0.8],
        ])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, 0.4, 0.6, 0.8])
        pd.testing.assert_series_equal(expected_output, output)

    def test_reverse_transform__null_column_false_nulls_true(self):
        """Test reverse_transform when _null_column is False and nulls.

        When _null_column is False and the nulls attribute, the
        values in the input data that match the _fill_value attribute are
        replaced by nans.

        Setup:
            - NullTransformer instance with _null_column set to False and nulls
              attribute set to True, and with the _fill_value set to a scalar.

        Input:
            - 1d numpy array with variate float values, containing the _fill_value
              among them.

        Expected Output:
            - pd.Series containing the same data as input, with the values that
              match the _fill_value replaced by nans.
        """
        # Setup
        transformer = NullTransformer()
        transformer._null_column = False
        transformer.nulls = True
        transformer._fill_value = 0.4
        input_data = np.array([0.0, 0.2, 0.4, 0.6])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, np.nan, 0.6])
        pd.testing.assert_series_equal(expected_output, output)

    def test_reverse_transform__null_column_false_nulls_false(self):
        """Test reverse_transform when _null_column is False and nulls.

        When _null_column is False and the nulls attribute is also False, the
        input data is not modified at all.

        Setup:
            - NullTransformer instance with _null_column and nulls attributes
              set to False, and the _fill_value set to a scalar value.

        Input:
            - 1d numpy array with variate float values, containing the _fill_value
              among them.

        Expected Output:
            - pd.Series containing the same data as input, without modification.
        """
        # Setup
        transformer = NullTransformer()
        transformer._null_column = False
        transformer.nulls = False
        transformer._fill_value = 0.4
        input_data = np.array([0.0, 0.2, 0.4, 0.6])

        # Run
        output = transformer.reverse_transform(input_data)

        # Assert
        expected_output = pd.Series([0.0, 0.2, 0.4, 0.6])
        pd.testing.assert_series_equal(expected_output, output)



"""Basic Integration tests for the BaseTransformer."""

import numpy as np
import pandas as pd

from rdt.transformers.base import BaseTransformer


def test_dummy_transformer_series_output():
    """Test a transformer that outputs a Series.

    This validates that a Transformer that is implemented to
    produce a Series object as the output of the `_transform`
    method works as expected.

    Setup:
        - A DummyTransformer that implements a _transform
          that produces a Series and a reverse_transform
          that takes a Series as input.

    Input:
        - A DataFrame with a single column of boolean values

    Expected behavior:
        - The data should be transformed into a DataFrame that contains
          a single float column made of 0.0s and 1.0s.
        - The transformed data should be able to reversed to
          re-produce the input data.
    """
    # Setup
    class DummyTransformer(BaseTransformer):

        INPUT_TYPE = 'boolean'
        OUTPUT_TYPES = {
            'value': 'float'
        }

        def _fit(self, data):
            pass

        def _transform(self, data):
            return data.astype(float)

        def _reverse_transform(self, data):
            return data.round() != 0

    # Run
    data = pd.DataFrame({
        'bool': [True, False, True, False]
    })

    transformer = DummyTransformer()
    transformed = transformer.fit_transform(data, 'bool')

    reverse = transformer.reverse_transform(transformed)

    # Assert
    expected_transform = pd.DataFrame({
        'bool.value': [1., 0., 1., 0.]
    })
    pd.testing.assert_frame_equal(expected_transform, transformed)
    pd.testing.assert_frame_equal(reverse, data)


def test_dummy_transformer_dataframe_output():
    """Test a transformer that outputs a DataFrame.

    This validates that a Transformer that is implemented to
    produce a DataFrame object as the output of the `_transform`
    method works as expected.

    Setup:
        - A DummyTransformer that implements a _transform
          that produces a DataFrame and a reverse_transform
          that takes a DataFrame as input.

    Input:
        - A DataFrame with a single column of boolean values

    Expected behavior:
        - The data should be transformed into a DataFrame that contains
          a float column made of 0.0s, 1.0s and -1.0s, and another float
          column that contains 0.0s and 1.0s indicating which values were
          null in the input.
        - The transformed data should be able to reversed to
          re-produce the input data.
    """
    # Setup
    class DummyTransformer(BaseTransformer):

        INPUT_TYPE = 'boolean'
        OUTPUT_TYPES = {
            'value': 'float',
            'null': 'float'
        }

        def _fit(self, data):
            pass

        def _transform(self, data):
            out = pd.DataFrame(dict(zip(
                self.output_columns,
                [
                    data.astype(float).fillna(-1),
                    data.isna().astype(float)
                ]
            )))

            return out

        def _reverse_transform(self, data):
            output = data[self.output_columns[0]]
            output = output.round().astype(bool).astype(object)
            output.iloc[data[self.output_columns[1]] == 1] = np.nan

            return output

    # Run
    data = pd.DataFrame({
        'bool': [True, False, True, np.nan]
    })

    transformer = DummyTransformer()
    transformed = transformer.fit_transform(data, 'bool')

    reverse = transformer.reverse_transform(transformed)

    # Assert
    expected_transform = pd.DataFrame({
        'bool.value': [1., 0., 1., -1.],
        'bool.null': [0., 0., 0., 1.]
    })
    pd.testing.assert_frame_equal(expected_transform, transformed)
    pd.testing.assert_frame_equal(reverse, data)


def test_dummy_transformer_multi_column_input():
    """Test a transformer that inputs a DataFrame.

    This validates that a Transformer that is implemented to
    expect multiple columns as the input for the `_transform`
    method works as expected.

    Setup:
        - A DummyTransformer that implements a _transform
          that expects a DataFrame with multiple columns as
          input, and a reverse transform that produces the
          same set of columns.

    Input:
        - A DataFrame with three year, month and day columns that
          represent dates.

    Expected behavior:
        - The data should be transformed into a DataFrame that contains
          a single float column with the timestamp values of the inputed
          datetimes.
        - The data should be able to be transformed and reverse
          transformed to re-produce the input data.
    """
    # Setup
    class DummyTransformer(BaseTransformer):

        INPUT_TYPE = 'datetime'
        OUTPUT_TYPES = {
            'value': 'float',
        }

        def _fit(self, data):
            pass

        def _transform(self, data):
            # Convert multiple columns into a single datetime
            data = pd.to_datetime(data)

            float_data = data.to_numpy().astype(float)
            data_is_nan = data.isna().to_numpy().astype(float)

            output = dict(zip(
                self.output_columns,
                [float_data, data_is_nan]
            ))

            output = pd.DataFrame(output).fillna(-1)

            return output

        def _reverse_transform(self, data):
            datetimes = data.round().astype('datetime64[ns]')
            out = pd.DataFrame({
                'year': datetimes.dt.year,
                'month': datetimes.dt.month,
                'day': datetimes.dt.day,
            })

            return out

    # Run
    data = pd.DataFrame({
        'year': [2001, 2002, 2003],
        'month': [1, 2, 3],
        'day': [1, 2, 3],
    })

    transformer = DummyTransformer()
    transformed = transformer.fit_transform(data, ['year', 'month', 'day'])

    reverse = transformer.reverse_transform(transformed)

    # Assert
    expected_transform = pd.DataFrame({
        'year#month#day.value': [9.783072e+17, 1.012608e+18, 1.046650e+18]
    })
    pd.testing.assert_frame_equal(expected_transform, transformed)
    pd.testing.assert_frame_equal(reverse, data)



from rdt.transformers import BooleanTransformer, get_transformer_class, get_transformer_instance


def test_get_transformer_class_transformer_name():
    """Test the ``get_transformer_class`` method.

    Validate the method returns the correct class when passed the class name.

    Input:
        - a string describing the transformer name.

    Output:
        - the class corresponding to the transformer name.
    """
    # Setup
    transformer_name = 'BooleanTransformer'

    # Run
    returned = get_transformer_class(transformer_name)

    # Assert
    assert returned == BooleanTransformer


def test_get_transformer_class_transformer_path():
    """Test the ``get_transformer_class`` method.

    Validate the method returns the correct class when passed the class path.

    Input:
        - a string describing the transformer path.

    Output:
        - the class corresponding to the transformer path.
    """
    # Setup
    transformer_path = 'rdt.transformers.BooleanTransformer'

    # Run
    returned = get_transformer_class(transformer_path)

    # Assert
    assert returned == BooleanTransformer


def test_get_transformer_instance_instance():
    transformer = BooleanTransformer(nan=None)

    returned = get_transformer_instance(transformer)

    assert isinstance(returned, BooleanTransformer)
    assert returned.nan is None


def test_get_transformer_instance_str():
    transformer = 'BooleanTransformer'

    returned = get_transformer_instance(transformer)

    assert isinstance(returned, BooleanTransformer)


def test_get_transformer_instance_class():
    transformer = BooleanTransformer

    returned = get_transformer_instance(transformer)

    assert isinstance(returned, BooleanTransformer)



import numpy as np
import pandas as pd

from rdt.transformers.datetime import DatetimeTransformer


def test_no_strip():
    dtt = DatetimeTransformer(strip_constant=False)
    data = pd.to_datetime(pd.Series([None, '1996-10-17', '1965-05-23']))

    # Run
    dtt._fit(data.copy())
    transformed = dtt._transform(data.copy())
    reverted = dtt._reverse_transform(transformed)

    # Asserts
    expect_trans = np.array([
        [350006400000000000, 1.0],
        [845510400000000000, 0.0],
        [-145497600000000000, 0.0]
    ])
    np.testing.assert_almost_equal(expect_trans, transformed)
    pd.testing.assert_series_equal(reverted, data)


def test_strip():
    dtt = DatetimeTransformer(strip_constant=True)
    data = pd.to_datetime(pd.Series([None, '1996-10-17', '1965-05-23']))

    # Run
    dtt._fit(data.copy())
    transformed = dtt._transform(data.copy())
    reverted = dtt._reverse_transform(transformed)

    # Asserts
    expect_trans = np.array([
        [4051.0, 1.0],
        [9786.0, 0.0],
        [-1684.0, 0.0]
    ])
    np.testing.assert_almost_equal(expect_trans, transformed)
    pd.testing.assert_series_equal(reverted, data)



from unittest.mock import Mock, patch

import numpy as np
import pandas as pd
import pytest

from rdt.transformers.datetime import DatetimeRoundedTransformer, DatetimeTransformer
from rdt.transformers.null import NullTransformer


class TestDatetimeTransformer:

    def test___init__(self):
        """Test the ``__init__`` method.

        Validate the passed arguments are stored as attributes.

        Setup:
            - initialize a ``DatetimeTransformer`` with values for each parameter.

        Side effect:
            - the ``nan`` attribute has been assigned as ``'mode'``.
            - the ``null_column`` attribute has been assigned as True.
            - the ``strip_constant`` attribute has been assigned as True.
        """
        # Setup
        transformer = DatetimeTransformer(
            nan='mode',
            null_column=True,
            strip_constant=True
        )

        # Asserts
        assert transformer.nan == 'mode'
        assert transformer.null_column is True
        assert transformer.strip_constant is True

    def test_is_composition_identity_null_transformer_true(self):
        """Test the ``is_composition_identity`` method with a ``null_transformer``.

        When the attribute ``null_transformer`` is not None and a null column is not created,
        this method should simply return False.

        Setup:
            - initialize a ``DatetimeTransformer`` transformer which sets
            ``self.null_transformer`` to a ``NullTransformer``.

        Output:
            - False.
        """
        # Setup
        transformer = DatetimeTransformer()
        transformer.null_transformer = NullTransformer(fill_value='fill')

        # Run
        output = transformer.is_composition_identity()

        # Assert
        assert output is False

    def test_is_composition_identity_null_transformer_false(self):
        """Test the ``is_composition_identity`` method without a ``null_transformer``.

        When the attribute ``null_transformer`` is None, this method should return
        the value stored in the ``COMPOSITION_IS_IDENTITY`` attribute.

        Setup:
            - initialize a ``DatetimeTransformer`` transformer which sets
            ``self.null_transformer`` to None.

        Output:
            - the value stored in ``self.COMPOSITION_IS_IDENTITY``.
        """
        # Setup
        transformer = DatetimeTransformer()
        transformer.null_transformer = None

        # Run
        output = transformer.is_composition_identity()

        # Assert
        assert output is True

    def test_get_output_types(self):
        """Test the ``get_output_types`` method when a null column is created.

        When a null column is created, this method should apply the ``_add_prefix``
        method to the following dictionary of output types:

        output_types = {
            'value': 'float',
            'is_null': 'float'
        }

        Setup:
            - initialize a ``DatetimeTransformer`` transformer which:
                - sets ``self.null_transformer`` to a ``NullTransformer`` where
                ``self._null_column`` is True.
                - sets ``self.column_prefix`` to a column name.

        Output:
            - the ``output_types`` dictionary, but with the ``self.column_prefix``
            added to the beginning of the keys.
        """
        # Setup
        transformer = DatetimeTransformer()
        transformer.null_transformer = NullTransformer(fill_value='fill')
        transformer.null_transformer._null_column = True
        transformer.column_prefix = 'a#b'

        # Run
        output = transformer.get_output_types()

        # Assert
        expected = {
            'a#b.value': 'float',
            'a#b.is_null': 'float'
        }
        assert output == expected

    def test__find_divider(self):
        """Test the ``_find_divider`` method.

        Find the greatest common denominator out of these values: [10] * 9 + [60, 60, 24],
        where each consecutive value in the list is multiplied by the previous one
        (so 10, 100, 1000, etc).

        Input:
            - a numpy array.

        Side effect:
            - sets ``self.divider`` to the correct divider.
        """
        # Setup
        data = np.array([100, 7919])
        transformer = DatetimeTransformer()

        # Run
        transformer._find_divider(data)

        # Assert
        assert transformer.divider == 1

    def test__convert_to_datetime(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure the transformer converts the data to datetime
        if it is of type ``object`` and can be converted.

        Input:
            - a pandas Series of dtype object, with elements that can be
            converted to datetime.

        Output:
            - a pandas series of type datetime.
        """
        # Setup
        data = pd.Series(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = DatetimeTransformer()

        # Run
        converted_data = transformer._convert_to_datetime(data)

        # Assert
        expected_data = pd.Series(pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01']))
        pd.testing.assert_series_equal(expected_data, converted_data)

    def test__convert_to_datetime_format(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure the transformer uses the specified format
        to convert the data to datetime.

        Setup:
            - The transformer will be initialized with a format.
        Input:
            - a pandas Series of dtype object, with elements in the specified
            format.

        Output:
            - a pandas series of type datetime.
        """
        # Setup
        data = pd.Series(['01Feb2020', '02Mar2020', '03Jan2010'])
        dt_format = '%d%b%Y'
        transformer = DatetimeTransformer(datetime_format=dt_format)

        # Run
        converted_data = transformer._convert_to_datetime(data)

        # Assert
        expected_data = pd.Series(pd.to_datetime(['01Feb2020', '02Mar2020', '03Jan2010']))
        pd.testing.assert_series_equal(expected_data, converted_data)

    def test__convert_to_datetime_not_convertible_raises_error(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure a ``TypeError`` is raised if the data is of type
        ``object`` but can't be converted.

        Input:
            - a pandas Series of dtype object, with elements that can't be
            converted to datetime.

        Expected behavior:
            - a ``TypeError`` is raised.
        """
        # Setup
        data = pd.Series(['2020-01-01-can', '2020-02-01-not', '2020-03-01-convert'])
        transformer = DatetimeTransformer()

        # Run
        error_message = 'Data must be of dtype datetime, or castable to datetime.'
        with pytest.raises(TypeError, match=error_message):
            transformer._convert_to_datetime(data)

    def test__convert_to_datetime_wrong_format_raises_error(self):
        """Test the ``_convert_to_datetime`` method.

        Test to make sure the transformer raises an error if the data does
        not match the specified format.

        Setup:
            - The transformer will be initialized with a format.
        Input:
            - a pandas Series of dtype object, with elements in the wrong
            format.

        Output:
            - a pandas series of type datetime.
        """
        # Setup
        data = pd.Series(['01-02-2020', '02-03-2020', '03J-01-2010'])
        dt_format = '%d%b%Y'
        transformer = DatetimeTransformer(datetime_format=dt_format)

        # Run
        error_message = 'Data does not match specified datetime format.'
        with pytest.raises(ValueError, match=error_message):
            transformer._convert_to_datetime(data)

    def test__transform_helper_calls_convert_to_datetime(self):
        """Test the ``_transform_helper`` method.

        Validate the helper transformer produces the correct value with ``strip_constant`` True.

        Input:
            - a pandas series of datetimes.

        Output:
            - a pandas series of the transformed datetimes.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = DatetimeTransformer()
        transformer._convert_to_datetime = Mock()
        transformer._convert_to_datetime.return_value = data

        # Run
        transformer._transform_helper(data)

        # Assert
        transformer._convert_to_datetime.assert_called_once_with(data)

    def test__transform_helper_strip_constant_true(self):
        """Test the ``_transform_helper`` method.

        Validate the helper transformer produces the correct value with ``strip_constant`` True.

        Input:
            - a pandas series of datetimes.

        Output:
            - a pandas series of the transformed datetimes.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = DatetimeTransformer(strip_constant=True)

        # Run
        transformed = transformer._transform_helper(data)

        # Assert
        np.testing.assert_allclose(transformed, np.array([
            18262., 18293., 18322.,
        ]))

    def test__transform_helper_strip_constant_false(self):
        """Test the ``_transform_helper`` method.

        Validate the helper transformer produces the correct value with ``strip_constant`` False.

        Input:
            - a pandas series of datetimes.

        Output:
            - a pandas series of the transformed datetimes.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = DatetimeTransformer(strip_constant=False)

        # Run
        transformed = transformer._transform_helper(data)

        # Assert
        np.testing.assert_allclose(transformed, np.array([
            1.577837e+18, 1.580515e+18, 1.583021e+18,
        ]), rtol=1e-5)

    @patch('rdt.transformers.datetime.NullTransformer')
    def test__fit(self, null_transformer_mock):
        """Test the ``_fit`` method for numpy arrays.

        Validate that this method (1) sets ``self.null_transformer`` to the
        ``NullTransformer`` with the correct attributes, (2) calls the
        ``self.null_transformer``'s ``fit`` method, and (3) calls the
        ``_transform_helper`` method with the transformed data.

        Input:
            - a pandas Series.

        Side effects:
            - sets ``self.null_transformer`` to the ``NullTransformer`` with
            the correct attributes.
            - calls the ``self.null_transformer``'s ``fit`` method.
            - calls the ``_transform_helper`` method with the transformed data.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = DatetimeTransformer()

        # Run
        transformer._fit(data)

        # Assert
        null_transformer_mock.assert_called_once_with('mean', None, copy=True)
        assert null_transformer_mock.return_value.fit.call_count == 1
        np.testing.assert_allclose(
            null_transformer_mock.return_value.fit.call_args_list[0][0],
            np.array([[1.577837e+18, 1.580515e+18, 1.583021e+18]]), rtol=1e-5
        )

    def test__transform(self):
        """Test the ``_transform`` method for numpy arrays.

        Validate that this method calls the ``null_transformer.transform`` method.
        It should also check that the final output is correct.

        Setup:
            - mock behavior of ``_fit``.

        Input:
            - a pandas Series.

        Output:
            - a numpy array containing the transformed data.

        Side effect:
            - calls the ``null_transformer.transform`` method with the transformed data.
        """
        # Setup
        data = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        transformer = DatetimeTransformer()
        transformer.null_transformer = Mock()

        # Run
        transformer._transform(data)

        # Assert
        assert transformer.null_transformer.transform.call_count == 1
        np.testing.assert_allclose(
            transformer.null_transformer.transform.call_args_list[0][0],
            np.array([[1.577837e+18, 1.580515e+18, 1.583021e+18]]), rtol=1e-5
        )

    def test__reverse_transform_all_none(self):
        dt = pd.to_datetime(['2020-01-01'])
        dtt = DatetimeTransformer(strip_constant=True)
        dtt._fit(dt)

        output = dtt._reverse_transform(pd.Series([None]))

        expected = pd.Series(pd.to_datetime(['NaT']))
        pd.testing.assert_series_equal(output, expected)

    def test__reverse_transform_2d_ndarray(self):
        """Test the ``_reverese_transform`` method for 2d arrays.

        Validate that the method correctly reverse transforms 2d arrays.

        Input:
            - a numpy 2d array.

        Output:
            - a pandas Series of datetimes.
        """
        # Setup
        dt = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        dtt = DatetimeTransformer(nan=None, strip_constant=True)
        dtt._fit(dt)
        transformed = np.array([[18262.], [18293.], [18322.]])

        # Run
        output = dtt._reverse_transform(transformed)

        # Assert
        expected = pd.to_datetime(['2020-01-01', '2020-02-01', '2020-03-01'])
        pd.testing.assert_series_equal(output.to_series(), expected.to_series())


class TestDatetimeRoundedTransformer:

    def test___init___strip_is_true(self):
        """Test that by default the ``strip_constant`` is set to True."""
        dtrt = DatetimeRoundedTransformer()

        # assert
        assert dtrt.strip_constant
